unit FontSubsetterTTFTools;

{
    This file is part of Synopse framework.

    Synopse framework. Copyright (C) 2022 Arnaud Bouchez
      Synopse Informatique - https://synopse.info

  *** BEGIN LICENSE BLOCK *****
  This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
  If a copy of the MPL was not distributed with this file, you can obtain one at http://mozilla.org/MPL/2.0/.
  This Source Code Form is “Incompatible With Secondary Licenses”, as defined by the Mozilla Public License, v. 2.0.
  
  Software distributed under the License is distributed on an "AS IS" basis,
  WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
  for the specific language governing rights and limitations under the License.

  The Original Code is Synopse framework.

  The Initial Developer of the Original Code is Arnaud Bouchez.

  Portions created by the Initial Developer are Copyright (C) 2022
  the Initial Developer. All Rights Reserved.

  Contributor(s):
   Achim Kalwa
   Alexander (chaa)
   aweste
   CoMPi
   Damien (ddemars)
   David Mead (MDW)
   David Heffernan
   FalconB
   Florian Grummel
   Harald Simon
   Josh Kelley (joshkel)
   Karel (vandrovnik)
   Kukhtin Igor
   LoukaO
   Marsh
   MChaos
   Mehrdad Momeni (nosa)
   mogulza
   Nzsolt
   Ondrej (reddwarf)
   Pierre le Riche
   Sinisa (sinisav)
   Sundazer

  ***** END LICENSE BLOCK *****

  Sponsors: https://synopse.info/fossil/wiki?name=HelpDonate
  Ongoing development and maintenance of the SynPDF library was sponsored
  in part by:
   https://www.helpndoc.com
    Easy to use yet powerful help authoring environment which can generate
    various documentation formats from a single source.
  Thanks for your contribution!

  This modified version is contributed by the 3D-Tool GmbH & Co. KG, http://www.3D-Tool.de
  (c) 2023, 3D-Tool GmbH & Co. KG, team@3D-Tool.de


}

//changes: 
//  - Create a new unit
//  - Extraction of procedure ReduceTTF + assiciated constants, records and tapes from SynPdf.pas
//  - Adding of helper functions from SynCommons.pas and reduction to windows systems support in Delphi
//  - Change of function call 
//       from: procedure ReduceTTF(out ttf: PDFString; SubSetData: pointer; SubSetSize: integer);
//         to: procedure ReduceTTF(out SubSetDataReduced: PAnsiChar; out SubSetSizeReduced: Cardinal; SubSetData: PAnsiChar; SubSetSize: Integer);
//  - Adaptions to support the new calling parameter
//  - Code formatting changes
//  - No functionality change

interface

uses
  Winapi.Windows, Winapi.Messages,
  System.SysUtils, System.Classes, System.IOUtils,
  Vcl.Forms, Vcl.Graphics;

procedure ReduceTTF(out SubSetDataReduced: PAnsiChar; out SubSetSizeReduced: Cardinal; SubSetData: PAnsiChar; SubSetSize: Integer);

implementation

type
  PtrInt = {$IFDEF UNICODE}NativeInt{$ELSE}Integer{$ENDIF};
  /// a CPU-dependent signed integer type cast of a pointer of pointer
  // - used for 64 bits compatibility, native under Free Pascal Compiler
  PPtrInt = ^PtrInt;

/// 'head' table contains global information about the font
  TCmapHEAD = packed record
    version: longint;
    fontRevision: longint;
    checkSumAdjustment: Cardinal;
    magicNumber: Cardinal;
    flags: Word;
    unitsPerEm: Word;
    createdDate: Int64;
    modifiedDate: Int64;
    xMin: SmallInt;
    yMin: SmallInt;
    xMax: SmallInt;
    yMax: SmallInt;
    macStyle: Word;
    lowestRec: Word;
    fontDirection: SmallInt;
    indexToLocFormat: SmallInt;
    glyphDataFormat: SmallInt
  end;
  /// header for the 'cmap' Format 4 table
  // - this is a two-byte encoding format
  TCmapFmt4 = packed record
    Format: Word;
    Length: Word;
    language: Word;
    segCountX2: Word;
    searchRange: Word;
    entrySelector: Word;
    rangeShift: Word;
  end;

  TTtfTableDirectory = packed record
    sfntVersion: Cardinal; // 0x00010000 for version 1.0
    numTables: Word; // number of tables
    searchRange: Word; // HighBit(NumTables) x 16
    entrySelector: Word; // Log2(HighBit(NumTables))
    rangeShift: Word; // NumTables x 16 - SearchRange
  end;
  PTtfTableDirectory = ^TTtfTableDirectory;

  TTtfTableEntry = packed record
    Tag: Cardinal; // table identifier
    checksum: Cardinal; // checksum for this table
    offset: Cardinal; // offset from start of font file
    Length: Cardinal; // length of this table
  end;
  PTtfTableEntry = ^TTtfTableEntry;

  PtrUInt = NativeUInt;

  TCardinalArray = array[0..MaxInt div SizeOf(Cardinal) - 1] of Cardinal;
  PCardinalArray = ^TCardinalArray;

  TStrLen = longint;
  /// internal pointer integer type used for string header length field
  PStrLen = ^TStrLen;

const
  // see http://www.4real.gr/technical-documents-ttf-subset.html and
  // https://developer.apple.com/fonts/TrueType-Reference-Manual/RM06/Chap6.html
  TTF_SUBSET: array[0..9] of array[0..3] of AnsiChar = (
    'head', 'cvt ', 'fpgm', 'prep', 'hhea', 'maxp', 'hmtx', 'cmap', 'loca', 'glyf');

  HEAD_TABLE = $64616568; // 'head'
  TTCF_TABLE = $66637474; // 'ttcf'

  _STRLEN = SizeOf(TStrLen);

{$IFDEF CPUX64}
function bswap32(a: Cardinal): Cardinal;
asm .noframe
  mov eax, a
  bswap eax
end;
{$ENDIF}

{$IFDEF CPUX86}
function bswap32(a: Cardinal): Cardinal;
asm
  bswap eax
end;

{$ENDIF}

var MoveFast: procedure(const Source; var Dest; Count: PtrInt) = system.Move;

function IntegerScan(P: PCardinalArray; Count: PtrInt; Value: Cardinal): PCardinal;
begin
  Result:= nil;
  if P = nil then
    Exit;
  Count:= PtrInt(@P[Count - 4]);
  repeat
    if PtrUInt(P) > PtrUInt(Count) then
      Break;
    if P^[0] <> Value then
      if P^[1] <> Value then
        if P^[2] <> Value then
          if P^[3] <> Value then begin
            P:= @P[4];
            continue;
          end else
          Result:= @P[3] else
        Result:= @P[2] else
      Result:= @P[1] else
      Result:= Pointer(P);
    Exit;
  until False;
  Inc(Count, 4 * SizeOf(Value));
  Result:= Pointer(P);
  repeat
    if PtrUInt(Result) >= PtrUInt(Count) then
      Break;
    if Result^ = Value then
      Exit else
      Inc(Result);
  until False;
  Result:= nil;
end;

procedure ReduceTTF(out SubSetDataReduced: PAnsiChar; out SubSetSizeReduced: Cardinal; SubSetData: PAnsiChar; SubSetSize: Integer);
var dir: PTtfTableDirectory;
  d, e: PTtfTableEntry;
  head: ^TCmapHEAD;
  n, i, len: PtrInt;
  checksum: Cardinal;
begin
  GetMem(SubSetDataReduced, SubSetSize); // maximum size
  d:= Pointer(SubSetDataReduced);
  Inc(PTtfTableDirectory(d));
  // identify the tables to be included
  e:= PTtfTableEntry(SubSetData);
  Inc(PTtfTableDirectory(e));
  n:= 0;
  if SubSetSize > SizeOf(dir^) then
    for i:= 1 to swap(PTtfTableDirectory(SubSetData)^.numTables) do begin
      if IntegerScan(@TTF_SUBSET, Length(TTF_SUBSET), e^.Tag) <> nil then begin
        d^:= e^;
        Inc(d);
        Inc(n);
      end;
      Inc(e);
    end;
  // update the main directory
  if n < 8 then begin // pdf expects 10, and 8..15 for our fixed dir^ values
    MoveFast(SubSetData^, SubSetDataReduced^, SubSetSize); // paranoid
    SubSetSizeReduced := SubSetSize;
    Exit;
  end;
  dir:= PTtfTableDirectory(SubSetDataReduced);
  dir^.sfntVersion:= PTtfTableDirectory(SubSetData)^.sfntVersion;
  dir^.numTables:= swap(Word(n));
  //len := HighBit(n); // always 8 when n in 8..15
  //dir^.searchRange := swap(len * 16);
  //dir^.entrySelector := swap(Floor(log2(len))); // requires the Math unit
  //dir^.rangeShift := swap((integer(n) - len) * 16);
  dir^.searchRange:= 32768; // pre-computed values for n in 8..15
  dir^.entrySelector:= 768;
  dir^.rangeShift:= 8192;
  // include the associated data
  checksum:= 0;
  head:= nil;
  e:= PTtfTableEntry(SubSetDataReduced);
  Inc(PTtfTableDirectory(e));
  for i:= 1 to n do begin
    len:= bswap32(e^.Length);
    MoveFast(PByteArray(SubSetData)[bswap32(e^.offset)], d^, len);
    e^.offset:= bswap32(PtrUInt(d) - PtrUInt(SubSetDataReduced));
    if e^.Tag = HEAD_TABLE then // 'head' table
      head:= Pointer(d);
    while len and 3 <> 0 do begin // 32-bit padding
      PByteArray(d)[len]:= 0;
      Inc(len);
    end;
    Inc(checksum, bswap32(e^.checksum)); // we didn't change the table itself
    Inc(PByte(d), len);
    Inc(e);
  end;
  // finalize the generated content
  for i:= 0 to ((SizeOf(dir^) + (Integer(n) * SizeOf(e^))) shr 2) - 1 do
    Inc(checksum, PCardinalArray(SubSetDataReduced)[i]);
  if head <> nil then
    head^.checkSumAdjustment:= bswap32($B1B0AFBA - checksum);
  SubSetSizeReduced:= PtrUInt(d) - PtrUInt(SubSetDataReduced);
end;

end.

