{* ***** BEGIN LICENSE BLOCK *****
*
* This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
* If a copy of the MPL was not distributed with this file, you can obtain one at http://mozilla.org/MPL/2.0/.
* This Source Code Form is Incompatible With Secondary Licenses, as defined by the Mozilla Public License, v. 2.0.
*
* Software distributed under the License is distributed on an
* "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
* implied. See the License for the specific language governing
* rights and limitations under the License.
*
* The Original Code is Graphics32
* The Initial Developer of the Original Code is Alex A. Denisov
* Portions created by the Initial Developer are Copyright (C) 2000-2007
* the Initial Developer. All Rights Reserved.
* https://github.com/benstof/graphics32/
*
* Obtained through: GR32_Lines GR32_Text and GR32_Objects: gr32_lines_ver4.zip
* http://angusj.com/delphi/gr32_lines.php
*
* The Initial Developer of the code in GR32_Text.pas is Angus Johnson
* <angus@angusj.com>. GR32_Text.pas code is Copyright (C) 2009-2010.
* All Rights Reserved.
*
* Acknowledgements:
* Sub-pixel rendering based on code provided by Akaiten <akaiten@mail.ru>
* TLCDDistributionLut class provided by Akaiten and is based on code from
* Maxim Shemanarev.
*
* This modified version is contributed by 3D-Tool GmbH & Co. KG, https://www.3d-tool.de
* (c) Copyright 2018, 3D-Tool GmbH & Co. KG, team@3d-tool.de
*
* ***** END LICENSE BLOCK ***** *}


unit Gr32_Ex;

interface

uses Windows, Classes, Graphics, SysUtils;

type
  // TFloat determines the precision level for certain floating-point operations
  PFloat = ^TFloat;
  TFloat = Single;

  PFloatPoint = ^TFloatPoint;
  TFloatPoint = record
    X, Y: TFloat;
  end;

  // This type has data bits arrangement compatible with Windows.TFixed
  PFixed = ^TFixed;
  TFixed = type Integer;

  PFixedPoint = ^TFixedPoint;
  TFixedPoint = record
    X, Y: TFixed;
  end;
  PFixedPointArray = ^TFixedPointArray;

  TFixedPointArray = array[0..0] of TFixedPoint;
  PArrayOfFixedPoint = ^TArrayOfFixedPoint;
  TArrayOfFixedPoint = array of TFixedPoint;
  PArrayOfArrayOfFixedPoint = ^TArrayOfArrayOfFixedPoint;
  TArrayOfArrayOfFixedPoint = array of TArrayOfFixedPoint;

function GetTTFontCharInfo(MemDC: HDC; wc: WideChar; dx, dy: Single; mat2: TMat2; Hinted: Boolean;
  out gm: TGlyphMetrics; out polyPts: TArrayOfArrayOfFixedPoint): Boolean;
function GetMemoryDeviceContext(FontHandle: HFont; out MemDC: HDC; out oldFntHdl: HFont): Boolean;
procedure DeleteMemoryDeviceContext(MemDC: HDC; oldFntHdl: HFont);

implementation

const
  qbezier_tolerance = 0.5;
  cbezier_tolerance = 0.5;
  half = 0.5;

//------------------------------------------------------------------------------

function Fixed(S: Single): TFixed;
begin
  Result:= Round(S * 65536);
end;

function OffsetPoint(const pt: Windows.TPointfx; const dx, dy: Single): TFixedPoint;
begin
  Result:= TFixedPoint(pt);
  Result.X:= Result.X + Fixed(dx);
  Result.Y:= Result.Y + Fixed(dy);
end;

function FloatPoint(const FXP: TFixedPoint): TFloatPoint;
const F = 1 / 65536;
begin
  with FXP do
  begin
    Result.X:= X * F;
    Result.Y:= Y * F;
  end;
end;

function FixedPoint(const FP: TFloatPoint): TFixedPoint;
begin
  Result.X:= Round(FP.X * 65536);
  Result.Y:= Round(FP.Y * 65536);
end;

function MidPoint(pt1, pt2: TFixedPoint): TFixedPoint; overload;
begin
  Result.X:= (pt1.X + pt2.X) div 2;
  Result.Y:= (pt1.Y + pt2.Y) div 2;
end;

//------------------------------------------------------------------------------

function GetQSplinePoints(const control_points: array of TFixedPoint): TArrayOfFixedPoint;
var
  i, arrayLen, resultCnt, resultSize, arrayLenMin3: Integer;
  pts1, pts2, pts3: TFloatPoint;

  procedure RecursiveQSpline(const p1, p2, p3: TFloatPoint);
  var
    p12, p23, p123: TFloatPoint;
  begin
    //assess flatness of curve ...
    if Abs(p1.x + p3.x - 2 * p2.x) + Abs(p1.y + p3.y - 2 * p2.y) < qbezier_tolerance then
    begin
      if resultCnt = Length(Result) then SetLength(Result, Length(Result) + 128);
      Result[resultCnt]:= FixedPoint(p3);
      Inc(resultCnt);
    end else
    begin
      p12.X:= (p1.X + p2.X) * half;
      p12.Y:= (p1.Y + p2.Y) * half;
      p23.X:= (p2.X + p3.X) * half;
      p23.Y:= (p2.Y + p3.Y) * half;
      p123.X:= (p12.X + p23.X) * half;
      p123.Y:= (p12.Y + p23.Y) * half;
      RecursiveQSpline(p1, p12, p123);
      RecursiveQSpline(p123, p23, p3);
    end;
  end;

begin
  //QSplines are series of 'smooth' QBeziers where each shared point is also
  //the midpoint of the control points of the adjoining QBeziers. This function
  //is typically used together with the Windows' GetGlyphOutline() API function.
  //(c.f. SVG's 'T' command - http://www.w3.org/TR/SVG11/paths.html )

  //first check that the 'control_points' count is valid ...
  arrayLen:= Length(control_points);
  if (arrayLen < 3) then Exit;
  arrayLenMin3:= arrayLen - 3;
  resultSize:= 128;
  SetLength(Result, resultSize);
  Result[0]:= control_points[0];
  resultCnt:= 1;

  pts1:= FloatPoint(control_points[0]);
  for i:= 0 to arrayLenMin3 do
  begin
    pts2:= FloatPoint(control_points[i + 1]);
    if i < arrayLenMin3 then //if not last bezier then ...
      pts3:= FloatPoint(MidPoint(control_points[i + 1], control_points[i + 2]))
    else
      pts3:= FloatPoint(control_points[i + 2]);
    RecursiveQSpline(pts1, pts2, pts3);

    //prepare for the next segment ...
    pts1:= pts3;
  end;
  SetLength(Result, resultCnt);
end;

//------------------------------------------------------------------------------

function GetCBezierPoints(const control_points: array of TFixedPoint): TArrayOfFixedPoint;
var
  i, j, arrayLen, resultCnt: Integer;
  ctrlPts: array[0..3] of TFloatPoint;

  procedure RecursiveCBezier(const p1, p2, p3, p4: TFloatPoint);
  var
    p12, p23, p34, p123, p234, p1234: TFloatPoint;
  begin
    //assess flatness of curve ...
    //http://groups.google.com/group/comp.graphics.algorithms/tree/browse_frm/thread/d85ca902fdbd746e
    if Abs(p1.x + p3.x - 2 * p2.x) +
      Abs(p2.x + p4.x - 2 * p3.x) +
      Abs(p1.y + p3.y - 2 * p2.y) +
      Abs(p2.y + p4.y - 2 * p3.y) < cbezier_tolerance then
    begin
      if resultCnt = Length(Result) then
        SetLength(Result, Length(Result) + 128);
      Result[resultCnt]:= FixedPoint(p4);
      Inc(resultCnt);
    end else
    begin
      p12.X:= (p1.X + p2.X) * half;
      p12.Y:= (p1.Y + p2.Y) * half;
      p23.X:= (p2.X + p3.X) * half;
      p23.Y:= (p2.Y + p3.Y) * half;
      p34.X:= (p3.X + p4.X) * half;
      p34.Y:= (p3.Y + p4.Y) * half;
      p123.X:= (p12.X + p23.X) * half;
      p123.Y:= (p12.Y + p23.Y) * half;
      p234.X:= (p23.X + p34.X) * half;
      p234.Y:= (p23.Y + p34.Y) * half;
      p1234.X:= (p123.X + p234.X) * half;
      p1234.Y:= (p123.Y + p234.Y) * half;
      RecursiveCBezier(p1, p12, p123, p1234);
      RecursiveCBezier(p1234, p234, p34, p4);
    end;
  end;

begin
  //first check that the 'control_points' count is valid ...
  arrayLen:= Length(control_points);
  if (arrayLen < 4) or ((arrayLen - 1) mod 3 <> 0) then Exit;

  SetLength(Result, 128);
  Result[0]:= control_points[0];
  resultCnt:= 1;
  for i:= 0 to (arrayLen div 3) - 1 do
  begin
    for j:= 0 to 3 do
      ctrlPts[j]:= FloatPoint(control_points[i * 3 + j]);
    RecursiveCBezier(ctrlPts[0], ctrlPts[1], ctrlPts[2], ctrlPts[3]);
  end;
  SetLength(Result, resultCnt);
end;

//------------------------------------------------------------------------------

procedure ParseFontCharInfo(info, endInfo: PByte; dx, dy: Single; out polyPts: TArrayOfArrayOfFixedPoint);
const TT_PRIM_CSPLINE = $3;
var
  tmpCurvePts: TArrayOfFixedPoint;
  i, polyCnt, ptCnt, ptCntTot: Integer;
  endContour: PByte;
begin
  polyCnt:= -1;
  while Info < endInfo do
  begin
    with PTTPolygonHeader(info)^ do
    begin
      if dwType <> TT_POLYGON_TYPE then
        raise Exception.Create('GetGlyphOutline() error - wrong header type');
      endContour:= info + cb;
      Inc(info, SizeOf(TTTPolygonHeader));
      Inc(polyCnt);
      SetLength(polyPts, polyCnt + 1);
      SetLength(polyPts[polyCnt], 1);
      polyPts[polyCnt][0]:= OffsetPoint(pfxStart, dx, dy);
      ptCntTot:= 1;
    end;
    while info < endContour do
    begin
      with PTTPolyCurve(info)^ do
      begin
        ptCnt:= cpfx;
        Inc(info, SizeOf(Word) * 2);
        case wType of
          TT_PRIM_LINE:
            begin
              SetLength(polyPts[polyCnt], ptCntTot + ptCnt);
              for i:= 0 to ptCnt - 1 do
              begin
                polyPts[polyCnt][ptCntTot + i]:=
                  OffsetPoint(PPointfx(info)^, dx, dy);
                Inc(info, SizeOf(TPointfx));
              end;
            end;
          TT_PRIM_QSPLINE: //http://support.microsoft.com/kb/q87115/
            begin
              //tmpCurvePts is used for the QSPLINE control points ...
              SetLength(tmpCurvePts, ptCnt + 1);
              //must include the previous point in the QSPLINE ...
              tmpCurvePts[0]:= polyPts[polyCnt][ptCntTot - 1];
              for i:= 1 to ptCnt do
              begin
                tmpCurvePts[i]:= OffsetPoint(PPointfx(info)^, dx, dy);
                Inc(info, SizeOf(TPointfx));
              end;
              tmpCurvePts:= GetQSplinePoints(tmpCurvePts);
              ptCnt:= Length(tmpCurvePts) - 1; //nb: first point already added
              SetLength(polyPts[polyCnt], ptCntTot + ptCnt);
              Move(tmpCurvePts[1],
                polyPts[polyCnt][ptCntTot], ptCnt * SizeOf(TFixedPoint));
            end;
          TT_PRIM_CSPLINE:
            begin
              //tmpCurvePts is used for the CSPLINE control points ...
              SetLength(tmpCurvePts, ptCnt + 1);
              //must include the previous point in the CSPLINE ...
              tmpCurvePts[0]:= polyPts[polyCnt][ptCntTot - 1];
              for i:= 1 to ptCnt do
              begin
                tmpCurvePts[i]:= OffsetPoint(PPointfx(info)^, dx, dy);
                Inc(info, SizeOf(TPointfx));
              end;
              tmpCurvePts:= GetCBezierPoints(tmpCurvePts);
              ptCnt:= Length(tmpCurvePts) - 1; //nb: first point already added
              SetLength(polyPts[polyCnt], ptCntTot + ptCnt);
              Move(tmpCurvePts[1],
                polyPts[polyCnt][ptCntTot], ptCnt * SizeOf(TFixedPoint));
            end;
        else raise
          Exception.Create('GetGlyphOutline() error - wrong curve type');
        end;
        Inc(ptCntTot, ptCnt);
      end;
    end;
  end;
end;

//------------------------------------------------------------------------------

function GetTTFontCharInfo(MemDC: HDC; wc: WideChar; dx, dy: Single; mat2: TMat2; Hinted: Boolean;
  out gm: TGlyphMetrics; out polyPts: TArrayOfArrayOfFixedPoint): Boolean;
const GGO_UNHINTED = $100;
  hintBool: array[Boolean] of UINT = (GGO_UNHINTED, 0);
var
  Size: DWord;
  info, startInfo: PByte;
begin
  startInfo:= nil;
  Size:= Windows.GetGlyphOutlineW(memDC,
    Cardinal(wc), GGO_NATIVE or hintBool[Hinted], gm, 0, nil, mat2);
  Result:= (Size <> GDI_ERROR);
  if not Result or (Size = 0) then Exit;
  GetMem(info, Size);
  try
    startInfo:= info;
    if Windows.GetGlyphOutlineW(memDC, Cardinal(wc),
      GGO_NATIVE or hintBool[Hinted], gm, Size, info, mat2) <> GDI_ERROR then
      ParseFontCharInfo(info, info + Size, dx, dy, polyPts) else
      Result:= False;
  finally
    FreeMem(startInfo);
  end;
end;

//------------------------------------------------------------------------------

function GetMemoryDeviceContext(FontHandle: HFont; out MemDC: HDC; out oldFntHdl: HFont): Boolean;
begin
  Result:= False;
  if FontHandle = 0 then Exit;
  memDC:= Windows.CreateCompatibleDC(0);
  if memDC = 0 then Exit;
  oldFntHdl:= Windows.SelectObject(memDC, FontHandle);
  Result:= oldFntHdl <> 0;
  if not Result then DeleteDC(memDC);
end;

procedure DeleteMemoryDeviceContext(MemDC: HDC; oldFntHdl: HFont);
begin
  Windows.SelectObject(memDC, oldFntHdl);
  DeleteDC(memDC);
end;


end.

